@@ -51,13 +51,17 @@ from yd_commands.printing import (
     print_numbered_strings,
 )
 from yd_commands.property_names import *
-from yd_commands.settings import NAMESPACE_PREFIX_SEPARATOR
+from yd_commands.settings import (
+    NAMESPACE_PREFIX_SEPARATOR,
+    VAR_CLOSING_DELIMITER,
+    VAR_OPENING_DELIMITER,
+)
 from yd_commands.submit_utils import (
     UploadedFiles,
     format_yd_name,
     generate_task_input_list,
     pause_between_batches,
-    update_config_work_requirement,
+    update_config_work_requirement_object,
 )
 from yd_commands.type_check import (
     check_bool,
@@ -100,6 +104,14 @@ if ARGS_PARSER.dry_run:
 
 UPLOADED_FILES: Optional[UploadedFiles] = None
 
+# Names for environment variables that are automatically added
+# to the environment for each Task
+YD_WORK_REQUIREMENT_NAME = "YD_WORK_REQUIREMENT_NAME"
+YD_TASK_GROUP_NAME = "YD_TASK_GROUP_NAME"
+YD_TASK_GROUP_NUMBER = "YD_TASK_GROUP_NUMBER"
+YD_TASK_NAME = "YD_TASK_NAME"
+YD_TASK_NUMBER = "YD_TASK_NUMBER"
+
 
 @main_wrapper
 def main():
@@ -221,8 +233,10 @@ def submit_work_requirement(
     )
     # Lazy substitution of the Work Requirement name, now it's defined
     add_substitutions(subs={L_WR_NAME: ID})
+    # Re-process substitutions in the CONFIG_WR object
+    CONFIG_WR = update_config_work_requirement_object(CONFIG_WR)
+    # Re-process substitutions in the wr_data dictionary
     process_variable_substitutions_in_dict_insitu(wr_data)
-    CONFIG_WR = update_config_work_requirement(CONFIG_WR)
 
     # Handle any files that need to be uploaded
     global UPLOADED_FILES
@@ -321,6 +335,13 @@ def create_task_group(
     # Name the Task Group
     num_task_groups = len(wr_data[TASK_GROUPS])
     num_tasks = len(task_group_data[TASKS])
+    if num_tasks == 1:  # Account for Task expansion
+        num_tasks = check_int(
+            task_group_data.get(
+                TASK_COUNT, wr_data.get(TASK_COUNT, CONFIG_WR.task_count)
+            )
+        )
+
     # The following handles possible CSV substitution at the config.toml level
     try:
         if task_group_data.get(NAME, None) is None:
@@ -336,6 +357,17 @@ def create_task_group(
         )
     )
 
+    # Add lazy substitutions for use in any Task Group property
+    add_substitution_overwrite(L_TASK_COUNT, str(num_tasks))
+    add_substitution_overwrite(L_TASK_GROUP_NAME, task_group_name)
+    add_substitution_overwrite(
+        L_TASK_GROUP_NUMBER, formatted_number_str(tg_number, num_task_groups)
+    )
+    add_substitution_overwrite(L_TASK_GROUP_COUNT, str(num_task_groups))
+    process_variable_substitutions_in_dict_insitu(task_group_data)
+    # Create a copy of global CONFIG_WR and apply lazy substitutions
+    config_wr = update_config_work_requirement_object(deepcopy(CONFIG_WR))
+
     # Assemble the RunSpecification values for the Task Group;
     # 'task_types' can automatically be added to by the task_types
     # specified in the Tasks.
@@ -344,13 +376,13 @@ def create_task_group(
             check_list(task_group_data.get(TASK_TYPES, wr_data.get(TASK_TYPES, [])))
         ).union(task_types_from_tasks)
     )
-    if CONFIG_WR.task_type is not None and CONFIG_WR.task_type not in task_types:
-        task_types.append(CONFIG_WR.task_type)
+    if config_wr.task_type is not None and config_wr.task_type not in task_types:
+        task_types.append(config_wr.task_type)
     if len(task_types) == 0:
         raise Exception(f"No Task Type(s) specified in Task Group '{task_group_name}'")
 
     vcpus_data: Optional[List[float]] = check_list(
-        task_group_data.get(VCPUS, wr_data.get(VCPUS, CONFIG_WR.vcpus))
+        task_group_data.get(VCPUS, wr_data.get(VCPUS, config_wr.vcpus))
     )
     vcpus = (
         None
@@ -359,7 +391,7 @@ def create_task_group(
     )
 
     ram_data: Optional[List[float]] = check_list(
-        task_group_data.get(RAM, wr_data.get(RAM, CONFIG_WR.ram))
+        task_group_data.get(RAM, wr_data.get(RAM, config_wr.ram))
     )
     ram = (
         None
@@ -368,7 +400,7 @@ def create_task_group(
     )
 
     providers_data: Optional[List[str]] = check_list(
-        task_group_data.get(PROVIDERS, wr_data.get(PROVIDERS, CONFIG_WR.providers))
+        task_group_data.get(PROVIDERS, wr_data.get(PROVIDERS, config_wr.providers))
     )
     providers: Optional[List[CloudProvider]] = (
         None
@@ -380,41 +412,41 @@ def create_task_group(
         taskTypes=task_types,
         maximumTaskRetries=check_int(
             task_group_data.get(
-                MAX_RETRIES, wr_data.get(MAX_RETRIES, CONFIG_WR.max_retries)
+                MAX_RETRIES, wr_data.get(MAX_RETRIES, config_wr.max_retries)
             )
         ),
         workerTags=check_list(
             task_group_data.get(
-                WORKER_TAGS, wr_data.get(WORKER_TAGS, CONFIG_WR.worker_tags)
+                WORKER_TAGS, wr_data.get(WORKER_TAGS, config_wr.worker_tags)
             )
         ),
         exclusiveWorkers=check_bool(
             task_group_data.get(
                 EXCLUSIVE_WORKERS,
-                wr_data.get(EXCLUSIVE_WORKERS, CONFIG_WR.exclusive_workers),
+                wr_data.get(EXCLUSIVE_WORKERS, config_wr.exclusive_workers),
             )
         ),
         instanceTypes=check_list(
             task_group_data.get(
-                INSTANCE_TYPES, wr_data.get(INSTANCE_TYPES, CONFIG_WR.instance_types)
+                INSTANCE_TYPES, wr_data.get(INSTANCE_TYPES, config_wr.instance_types)
             )
         ),
         vcpus=vcpus,
         ram=ram,
-        minWorkers=check_int(task_group_data.get(MIN_WORKERS, CONFIG_WR.min_workers)),
-        maxWorkers=check_int(task_group_data.get(MAX_WORKERS, CONFIG_WR.max_workers)),
+        minWorkers=check_int(task_group_data.get(MIN_WORKERS, config_wr.min_workers)),
+        maxWorkers=check_int(task_group_data.get(MAX_WORKERS, config_wr.max_workers)),
         tasksPerWorker=check_int(
-            task_group_data.get(TASKS_PER_WORKER, CONFIG_WR.tasks_per_worker)
+            task_group_data.get(TASKS_PER_WORKER, config_wr.tasks_per_worker)
         ),
         providers=providers,
         regions=check_list(
-            task_group_data.get(REGIONS, wr_data.get(REGIONS, CONFIG_WR.regions))
+            task_group_data.get(REGIONS, wr_data.get(REGIONS, config_wr.regions))
         ),
     )
     ctttl_data = check_float_or_int(
         task_group_data.get(
             COMPLETED_TASK_TTL,
-            wr_data.get(COMPLETED_TASK_TTL, CONFIG_WR.completed_task_ttl),
+            wr_data.get(COMPLETED_TASK_TTL, config_wr.completed_task_ttl),
         )
     )
     completed_task_ttl = None if ctttl_data is None else timedelta(minutes=ctttl_data)
@@ -428,7 +460,7 @@ def create_task_group(
             task_group_data.get(
                 FINISH_IF_ALL_TASKS_FINISHED,
                 wr_data.get(
-                    FINISH_IF_ALL_TASKS_FINISHED, CONFIG_WR.finish_if_all_tasks_finished
+                    FINISH_IF_ALL_TASKS_FINISHED, config_wr.finish_if_all_tasks_finished
                 ),
             )
         ),
@@ -436,12 +468,12 @@ def create_task_group(
             task_group_data.get(
                 FINISH_IF_ANY_TASK_FAILED,
                 wr_data.get(
-                    FINISH_IF_ANY_TASK_FAILED, CONFIG_WR.finish_if_any_task_failed
+                    FINISH_IF_ANY_TASK_FAILED, config_wr.finish_if_any_task_failed
                 ),
             )
         ),
         priority=check_float_or_int(
-            task_group_data.get(PRIORITY, wr_data.get(PRIORITY, CONFIG_WR.priority))
+            task_group_data.get(PRIORITY, wr_data.get(PRIORITY, config_wr.priority))
         ),
         completedTaskTtl=completed_task_ttl,
     )
@@ -521,7 +553,6 @@ def add_tasks_to_task_group(
                 task_batch_size=TASK_BATCH_SIZE,
                 batch_number=batch_number,
                 num_tasks=num_tasks,
-                first_batch=batch_number == 0,
             )
 
         # Iterate through tasks in the batch
@@ -538,6 +569,7 @@ def add_tasks_to_task_group(
                     num_tasks,
                     tg_number,
                     num_task_groups,
+                    task_group.name,
                 )
             )
 
@@ -546,23 +578,24 @@ def add_tasks_to_task_group(
                 L_TASK_NUMBER, formatted_number_str(task_number, num_tasks)
             )
             process_variable_substitutions_in_dict_insitu(task)
+            config_wr = update_config_work_requirement_object(deepcopy(CONFIG_WR))
 
             executable = check_str(
                 task.get(
                     EXECUTABLE,
                     task_group_data.get(
-                        EXECUTABLE, wr_data.get(EXECUTABLE, CONFIG_WR.executable)
+                        EXECUTABLE, wr_data.get(EXECUTABLE, config_wr.executable)
                     ),
                 )
             )
             arguments_list = check_list(
                 task.get(
                     ARGS,
-                    wr_data.get(ARGS, task_group_data.get(ARGS, CONFIG_WR.args)),
+                    wr_data.get(ARGS, task_group_data.get(ARGS, config_wr.args)),
                 )
             )
             env = check_dict(
-                task.get(ENV, task_group_data.get(ENV, wr_data.get(ENV, CONFIG_WR.env)))
+                task.get(ENV, task_group_data.get(ENV, wr_data.get(ENV, config_wr.env)))
             )
 
             # Set up lists of files to input, verify
@@ -571,7 +604,7 @@ def add_tasks_to_task_group(
                     INPUTS_REQUIRED,
                     task_group_data.get(
                         INPUTS_REQUIRED,
-                        wr_data.get(INPUTS_REQUIRED, CONFIG_WR.inputs_required),
+                        wr_data.get(INPUTS_REQUIRED, config_wr.inputs_required),
                     ),
                 )
             )
@@ -580,7 +613,7 @@ def add_tasks_to_task_group(
                     VERIFY_AT_START,
                     task_group_data.get(
                         VERIFY_AT_START,
-                        wr_data.get(VERIFY_AT_START, CONFIG_WR.verify_at_start),
+                        wr_data.get(VERIFY_AT_START, config_wr.verify_at_start),
                     ),
                 )
             )
@@ -588,7 +621,7 @@ def add_tasks_to_task_group(
                 task.get(
                     VERIFY_WAIT,
                     task_group_data.get(
-                        VERIFY_WAIT, wr_data.get(VERIFY_WAIT, CONFIG_WR.verify_wait)
+                        VERIFY_WAIT, wr_data.get(VERIFY_WAIT, config_wr.verify_wait)
                     ),
                 )
             )
@@ -597,7 +630,7 @@ def add_tasks_to_task_group(
                     INPUTS_OPTIONAL,
                     task_group_data.get(
                         INPUTS_OPTIONAL,
-                        wr_data.get(INPUTS_OPTIONAL, CONFIG_WR.inputs_optional),
+                        wr_data.get(INPUTS_OPTIONAL, config_wr.inputs_optional),
                     ),
                 )
             )
@@ -625,7 +658,7 @@ def add_tasks_to_task_group(
                     UPLOAD_FILES,
                     task_group_data.get(
                         UPLOAD_FILES,
-                        wr_data.get(UPLOAD_FILES, CONFIG_WR.upload_files),
+                        wr_data.get(UPLOAD_FILES, config_wr.upload_files),
                     ),
                 )
             )
@@ -669,7 +702,7 @@ def add_tasks_to_task_group(
                         OUTPUTS_OPTIONAL,
                         task_group_data.get(
                             OUTPUTS_OPTIONAL,
-                            wr_data.get(OUTPUTS_OPTIONAL, CONFIG_WR.outputs_optional),
+                            wr_data.get(OUTPUTS_OPTIONAL, config_wr.outputs_optional),
                         ),
                     )
                 )
@@ -683,7 +716,7 @@ def add_tasks_to_task_group(
                         OUTPUTS_REQUIRED,
                         task_group_data.get(
                             OUTPUTS_REQUIRED,
-                            wr_data.get(OUTPUTS_REQUIRED, CONFIG_WR.outputs_required),
+                            wr_data.get(OUTPUTS_REQUIRED, config_wr.outputs_required),
                         ),
                     )
                 )
@@ -695,7 +728,7 @@ def add_tasks_to_task_group(
                     OUTPUTS_OTHER,
                     task_group_data.get(
                         OUTPUTS_OTHER,
-                        wr_data.get(OUTPUTS_OTHER, CONFIG_WR.outputs_other),
+                        wr_data.get(OUTPUTS_OTHER, config_wr.outputs_other),
                     ),
                 )
             ):
@@ -719,7 +752,7 @@ def add_tasks_to_task_group(
                     CAPTURE_TASKOUTPUT,
                     task_group_data.get(
                         CAPTURE_TASKOUTPUT,
-                        wr_data.get(CAPTURE_TASKOUTPUT, CONFIG_WR.capture_taskoutput),
+                        wr_data.get(CAPTURE_TASKOUTPUT, config_wr.capture_taskoutput),
                     ),
                 )
             ):
@@ -729,7 +762,7 @@ def add_tasks_to_task_group(
                 ALWAYS_UPLOAD,
                 task_group_data.get(
                     ALWAYS_UPLOAD,
-                    wr_data.get(ALWAYS_UPLOAD, CONFIG_WR.always_upload),
+                    wr_data.get(ALWAYS_UPLOAD, config_wr.always_upload),
                 ),
             )
 
@@ -746,20 +779,23 @@ def add_tasks_to_task_group(
                 if len(task_group.runSpecification.taskTypes) == 1:
                     task_type = task_group.runSpecification.taskTypes[0]
                 else:
-                    task_type = CONFIG_WR.task_type
+                    task_type = config_wr.task_type
 
             tasks_list.append(
                 create_task(
+                    config_wr=config_wr,
                     wr_data=wr_data,
                     task_group_data=task_group_data,
                     task_data=task,
                     task_name=task_name,
+                    task_number=task_number + 1,
                     tg_name=task_group.name,
+                    tg_number=tg_number + 1,
                     task_type=task_type,
                     executable=executable,
                     args=arguments_list,
                     task_data_property=get_task_data_property(
-                        wr_data, task_group_data, task, task_name
+                        config_wr, wr_data, task_group_data, task, task_name
                     ),
                     env=env,
                     inputs=inputs,
@@ -801,7 +837,11 @@ def add_tasks_to_task_group(
 
 
 def get_task_data_property(
-    wr_data: Dict, task_group_data: Dict, task: Dict, task_name: str
+    config_wr: ConfigWorkRequirement,
+    wr_data: Dict,
+    task_group_data: Dict,
+    task: Dict,
+    task_name: str,
 ) -> Optional[str]:
     """
     Get the 'taskData' property, either using the contents of the file
@@ -814,7 +854,7 @@ def get_task_data_property(
     for data, task_data_default, task_data_file_default in [
         (task, None, None),
         (task_group_data, None, None),
-        (wr_data, CONFIG_WR.task_data, CONFIG_WR.task_data_file),
+        (wr_data, config_wr.task_data, config_wr.task_data_file),
     ]:
         task_data_property = data.get(TASK_DATA, task_data_default)
         task_data_file_property = data.get(TASK_DATA_FILE, task_data_file_default)
@@ -865,8 +905,8 @@ def on_update(work_req: WorkRequirement):
         completed += task_group.taskSummary.statusCounts[TaskStatus.COMPLETED]
         total += task_group.taskSummary.taskCount
     print_log(
-        f"WORK REQUIREMENT is {work_req.status} with {completed}/{total} "
-        "COMPLETED TASKS"
+        f"Work Requirement is {work_req.status} with {completed}/{total} "
+        "completed Tasks"
     )
 
 
@@ -937,6 +977,7 @@ def get_task_name(
     num_tasks: int,
     task_group_number: int,
     num_task_groups: int,
+    task_group_name: str,
 ) -> str:
     """
     Create the name of a Task.
@@ -945,21 +986,25 @@ def get_task_name(
 
     if name:
         name = name.replace(
-            f"{{{{{L_TASK_NUMBER}}}}}",
+            f"{VAR_OPENING_DELIMITER + L_TASK_NUMBER + VAR_CLOSING_DELIMITER}",
             formatted_number_str(task_number, num_tasks),
         )
         name = name.replace(
-            f"{{{{{L_TASK_COUNT}}}}}",
+            f"{VAR_OPENING_DELIMITER + L_TASK_COUNT + VAR_CLOSING_DELIMITER}",
             str(num_tasks),
         )
         name = name.replace(
-            f"{{{{{L_TASK_GROUP_NUMBER}}}}}",
+            f"{VAR_OPENING_DELIMITER + L_TASK_GROUP_NUMBER + VAR_CLOSING_DELIMITER}",
             formatted_number_str(task_group_number, num_task_groups),
         )
         name = name.replace(
-            f"{{{{{L_TASK_GROUP_COUNT}}}}}",
+            f"{VAR_OPENING_DELIMITER + L_TASK_GROUP_COUNT + VAR_CLOSING_DELIMITER}",
             str(num_task_groups),
         )
+        name = name.replace(
+            f"{VAR_OPENING_DELIMITER + L_TASK_GROUP_NAME + VAR_CLOSING_DELIMITER}",
+            task_group_name,
+        )
 
     else:
         name = "task_" + formatted_number_str(task_number, num_tasks)
@@ -980,15 +1025,15 @@ def get_task_group_name(
 
     if name:
         name = name.replace(
-            f"{{{{{L_TASK_GROUP_NUMBER}}}}}",
+            f"{VAR_OPENING_DELIMITER + L_TASK_GROUP_NUMBER + VAR_CLOSING_DELIMITER}",
             formatted_number_str(task_group_number, num_task_groups),
         )
         name = name.replace(
-            f"{{{{{L_TASK_GROUP_COUNT}}}}}",
+            f"{VAR_OPENING_DELIMITER + L_TASK_GROUP_COUNT + VAR_CLOSING_DELIMITER}",
             str(num_task_groups),
         )
         name = name.replace(
-            f"{{{{{L_TASK_COUNT}}}}}",
+            f"{VAR_OPENING_DELIMITER + L_TASK_COUNT + VAR_CLOSING_DELIMITER}",
             str(task_count),
         )
 
@@ -998,12 +1043,18 @@ def get_task_group_name(
     return name
 
 
+PRINT_NO_EXECUTABLE_MESSAGE = True
+
+
 def create_task(
+    config_wr: ConfigWorkRequirement,
     wr_data: Dict,
     task_group_data: Dict,
     task_data: Dict,
     task_name: str,
+    task_number: int,
     tg_name: str,
+    tg_number: int,
     task_type: str,
     executable: str,
     args: List[str],
@@ -1038,9 +1089,6 @@ def create_task(
             taskData=task_data_property,
         )
 
-    check_list(args)
-    check_dict(env)
-
     # Flatten paths for downloaded files?
     flatten_input_paths: Optional[FlattenPath] = None
     if check_bool(
@@ -1048,28 +1096,31 @@ def create_task(
             FLATTEN_PATHS,
             task_group_data.get(
                 FLATTEN_PATHS,
-                wr_data.get(FLATTEN_PATHS, CONFIG_WR.flatten_input_paths),
+                wr_data.get(FLATTEN_PATHS, config_wr.flatten_input_paths),
             ),
         )
     ):
         flatten_input_paths = FlattenPath.FILE_NAME_ONLY
 
-    # Environment variable names
-    yd_task_name_str = "YD_TASK_NAME"
-    yd_tg_name_str = "YD_TASK_GROUP_NAME"
-    yd_wr_name_str = "YD_WORK_REQUIREMENT_NAME"
-
-    # Add names to the environment as a convenience
-    env_copy[yd_task_name_str] = task_name
-    env_copy[yd_tg_name_str] = tg_name
-    env_copy[yd_wr_name_str] = ID
+    # Add Task details to the environment as a convenience
+    if task_type != "docker":
+        env_copy[YD_TASK_NAME] = task_name
+        env_copy[YD_TASK_NUMBER] = str(task_number)
+        env_copy[YD_TASK_GROUP_NAME] = tg_name
+        env_copy[YD_TASK_GROUP_NUMBER] = str(tg_number)
+        env_copy[YD_WORK_REQUIREMENT_NAME] = ID
 
     # Special processing for Bash, Python & PowerShell tasks if the 'executable'
     # property is set. The script is uploaded if this hasn't already been done,
     # and added to the list of required files.
     if task_type in ["bash", "powershell", "python", "cmd", "bat"]:
         if executable is None:
-            print_log(f"Note: no 'executable' specified for '{task_type}' Task Type")
+            global PRINT_NO_EXECUTABLE_MESSAGE
+            if PRINT_NO_EXECUTABLE_MESSAGE:
+                print_log(
+                    f"Note: no 'executable' specified for '{task_type}' Task Type"
+                )
+                PRINT_NO_EXECUTABLE_MESSAGE = False
             return _make_task(flatten_input_paths)
 
         UPLOADED_FILES.add_input_file(
@@ -1116,14 +1167,17 @@ def create_task(
                 DOCKER_ENV,
                 task_group_data.get(
                     DOCKER_ENV,
-                    wr_data.get(DOCKER_ENV, CONFIG_WR.docker_env),
+                    wr_data.get(DOCKER_ENV, config_wr.docker_env),
                 ),
             )
         )
-        # Names are set in the container env. for convenience
-        docker_env_list = ["--env", f"{yd_task_name_str}={task_name}"]
-        docker_env_list += ["--env", f"{yd_tg_name_str}={tg_name}"]
-        docker_env_list += ["--env", f"{yd_wr_name_str}={ID}"]
+        # Add entity names to the container env. for convenience
+        docker_env_list = ["--env", f"{YD_TASK_NAME}={task_name}"]
+        docker_env_list += ["--env", f"{YD_TASK_NUMBER}={task_number}"]
+        docker_env_list += ["--env", f"{YD_TASK_GROUP_NAME}={tg_name}"]
+        docker_env_list += ["--env", f"{YD_TASK_GROUP_NUMBER}={tg_number}"]
+        docker_env_list += ["--env", f"{YD_WORK_REQUIREMENT_NAME}={ID}"]
+
         if docker_env is not None:
             for key, value in docker_env.items():
                 docker_env_list += ["--env", f"{key}={value}"]
@@ -1136,14 +1190,14 @@ def create_task(
             DOCKER_USERNAME,
             task_group_data.get(
                 DOCKER_USERNAME,
-                wr_data.get(DOCKER_USERNAME, CONFIG_WR.docker_username),
+                wr_data.get(DOCKER_USERNAME, config_wr.docker_username),
             ),
         )
         docker_password = task_data.get(
             DOCKER_PASSWORD,
             task_group_data.get(
                 DOCKER_PASSWORD,
-                wr_data.get(DOCKER_PASSWORD, CONFIG_WR.docker_password),
+                wr_data.get(DOCKER_PASSWORD, config_wr.docker_password),
             ),
         )
         env_copy.update(
